# VDaq analysis package.

__version__ = "0.8.5.dev1"

from .GTimer import GTimer
from .VDaqData import VDaqData
from .Progress import ShowProgress
from .AliVATAModule import AliVATAModule
from .ScanManager import ScanManager

def analyze_data():
    """Analyze data."""
    from .cmmds.analyze_data import analyze_data
    analyze_data()

def show_data():
    """Analyze data."""
    from .cmmds.show_data import show_data
    show_data()

def getSpectrum():
    """Makes an spectrum."""
    from .cmmds.getSpectrum import getSpectrum
    getSpectrum()

def getFileInfo():
    """Makes an spectrum."""
    from .cmmds.getFileInfo import getFileInfo
    getFileInfo()

def variableScan():
    """Show variation of signal in a scan."""
    from .cmmds.variable_scan import main
    main()




#!/usr/bin/env python3
"""A strip cluster."""

class Cluster:
    """This represents a strip data cluster.

    Members:
        chan: the "CoG" channel
        seed: the seed channel
        e_seed: the energy of the seed channel
        E: total energy
        first: first channel
        last: last channel.

    """
    def __init__(self, chan=-1.0, E=0.0):
        """initialization"""
        self.E = 0
        self.chan = -1
        self.seed = -1
        self.e_seed = 0.0
        self.first = 9999999
        self.last = -1
        
        if chan >=0:
            self.add_seed(chan, E)


    def add_seed(self, chan, E):
        """Adds the  seed."""
        self.seed = chan
        self.e_seed = E
        self.add(chan, E)
        
    def add(self, chan, E):
        """Adds a new channel."""
        if self.E > 0:
            self.chan = ( self.chan * self.E + chan*E )/(self.E + E)
        else:
            self.chan = chan
            
        self.E += E
        self.first = min(chan, self.first)
        self.last = max(chan, self.last)
        
        

    def set_from_data(self, channels, data):
        """Sets from data-"""
        self.E = 0.0
        self.seed = channels[0]
        self.chan = channels[0]
        channels = channels.sorted()
        self.first = channels[0]
        self.last = channels[-1]
        for i in channels:
            if data[i] > 0.0:
                self.E += data[i]
                self.chan += i*data[i]

        if self.E > 0:
            self.chan /= self.E


#!/usr/bin/env python3
"""Python script to show the data in a file."""
import datetime
import os
import sys
from argparse import Action
from argparse import ArgumentParser

import matplotlib.pyplot as plt
import numpy as np
from vdaq_ana import VDaqData
from vdaq_ana import AliVATAModule
from matplotlib.widgets import Button


def do_show_data(fname, options):
    """The main function."""
    # Check that the file exists
    if not os.path.exists(fname):
        print("Input file", fname, "does not exist")
        return

    # We open here the file with VDaqData
    vdaq = VDaqData(fname)
    vdaq.show_info(show_modules=True)

    n_mod = len(vdaq.modules)
    fig, ax = plt.subplots(nrows=1, ncols=n_mod)
    fig.set_figwidth(2.5*n_mod)
    dax = {}
    try:
        for a, m in zip(ax, vdaq.modules.values()):
            dax[m.id] = a

    except TypeError:
        m = list(vdaq.modules.values())[0]
        dax[m.id] = ax

    mod_evt = {}
    mod_dtime = {}
    for m in vdaq.modules.values():
        mod_evt[m.id] = 0.0
        mod_dtime[m.id] = 0.0

        m.seed_cut = options.seed_cut
        m.neigh_cut = options.neigh_cut

    plt.subplots_adjust(bottom=0.2)

    if options.scan_point is not None:
        vdaq_iter = vdaq.create_iterator_at_scanpoint(options.scan_point)
    elif options.start_time is not None:
        vdaq_iter = vdaq.create_iterator_at_time(options.start_time)
    elif options.start_evt is not None:
        vdaq_iter = vdaq.create_iterator_at_event(options.start_evt)
    else:
        vdaq_iter = iter(vdaq)

    def next_event(dummy):
        evt = next(vdaq_iter)
        mid = evt.mod_id
        md = vdaq.modules[mid]
        nbin = md.pedestal.shape[0]
        nchan = evt.data.shape[0]
        x = np.arange(0, nbin, 1)

        # Get the indices and build up
        if evt.romode == AliVATAModule.SPARSE_ADJ:
            indx = [i + evt.chan for i in md.adjacents]
            values = np.zeros(nbin)
            try:
                values[indx] = evt.data

            except IndexError:
                print("Bin {} out of bounds for module {}".format(indx, mid))

        elif evt.romode == AliVATAModule.SERIAL:
            values = np.array(evt.data, dtype='d')

        else:
            print("Readout mode not supported:", evt.romode)

        if options.pedestals:
            if evt.romode == AliVATAModule.SERIAL:
                values -= md.pedestal
                sn = values/md.noise
                cmmd = np.mean(values[np.nonzero(abs(sn) < 5.0)])
                values -= cmmd

            else:
                values[indx] -= md.pedestal[indx]
                sn = values/md.noise
                cmmd = np.mean(values[np.nonzero(abs(sn) < 5.0)])
                values[indx] -= cmmd


        dt = evt.time-mod_dtime[mid]
        mod_dtime[mid] = evt.time
        out = md.process_event(evt)
        clst = "".join(["({:>3d} {:>6.1f}) ".format(int(C.chan), C.E) for C in out])
        print("Module {}: daq_time {:>10.6f} evt no. {:>6d} evt d_time {:>10.3f} clusters {}".format(mid,
                                                                                                     evt.daq_time/1.e6,
                                                                                                     evt.evtcnt,
                                                                                                     dt*0.025,
                                                                                                     clst))

        ax = dax[mid]
        ax.clear()
        ax.set_title('Module %d' % mid)
        ax.step(x, values, where='mid')
        if options.fix_range:
            if options.pedestals:
                ax.set_ylim(-250, 2048)
            else:
                ax.set_ylim([0, 4096])

        ax.set_ymargin(0.2)
        ymin, ymax = ax.get_ylim()
        xmin, xmax = ax.get_xlim()
        dy = (ymax-ymin)
        dx = (xmax-xmin)
        ax.text(xmin+0.05*dx, ymax-0.05*dy, "Evt. {}".format(evt.evtcnt) )

        plt.draw()

    def loop_events(dummy):
        while options.plot:
            next_event(None)
            plt.draw()
            plt.pause(0.01)

        plt.show()

    axnext = plt.axes([0.81, 0.05, 0.1, 0.075])
    bnext = Button(axnext, 'Next')
    bnext.on_clicked(next_event)

    # axloop = plt.axes([0.4, 0.05, 0.2, 0.075])
    # bloop = Button(axloop, 'Loop/Step')
    # bloop.on_clicked(loop_events)

    axquit = plt.axes([0.1, 0.05, 0.1, 0.075])
    bquit = Button(axquit, 'Quit')
    bquit.on_clicked(sys.exit)

    if options.plot:
        loop_events(None)

    next_event(None)
    plt.show()


class CommaSeparatedListAction(Action):
    """Create a list from the comma sepparated numbers at imput."""

    def __call__(self, parser, namespace, values, option_string=None):
        """The actual action."""
        setattr(namespace, self.dest, list(map(int, values.split(','))))


def show_data():
    """Define arguments and call main function."""

    parser = ArgumentParser()
    parser.add_argument('files', nargs='+', help="Input files")
    parser.add_argument("--pedestals",
                        dest="pedestals", action="store_true",
                        help="Remove pedestals on the plot",
                        default=False)
    parser.add_argument("--fix_range", action="store_true",
                        help="Histogram range is fixed.",
                        default=False)
    parser.add_argument("--start-time", dest="start_time", default=None, type=float,
                        help="Event time to start showing")
    parser.add_argument("--start-evt", dest="start_evt", default=None,
                        action=CommaSeparatedListAction,
                        help="Event number to start showing")
    parser.add_argument("--scan-point", dest="scan_point", type=int, default=None,
                        help="Scan point number to visit")
    parser.add_argument("--plot", dest="plot", default=False,
                        action="store_true",
                        help="Draw histograms")
    parser.add_argument("--seed_cut", dest="seed_cut", default=3.5, type=float,
                        help="Threshold of charge for a cluster seed")
    parser.add_argument("--neigh_cut", dest="neigh_cut", default=1.5, type=float,
                        help="Threshold of charge to add a neighbour to a cluster")
    parser.add_argument("--raw-data", dest="raw_data",
                        default=False, action="store_true",
                        help="Show the raw data instead of the 'digested' signal")

    options = parser.parse_args()

    if len(options.files) == 0:
        print("I need an input file")
        sys.exit()

    do_show_data(options.files[0], options)


if __name__ == "__main__":
    show_data()


#!/usr/bin/env python3
"""Utilities."""

import numpy as np


def remove_outliers(data, cut=2.0, outliers=False, debug=False):
    """Remove points far away form the rest.

    Args:
    ----
        data : The data
        cut: max allowed distance
        outliers: if True, return the outliers rhater than remove them
        debug: be verbose if True.

    """
    d = np.abs(data - np.median(data))
    mdev = np.median(d)
    s = d / (mdev if mdev else 1.)
    if outliers:
        indx = np.where(s > cut)[0]
    else:
        indx = np.where(s < cut)[0]

    return indx


#!/usr/bin/env python3
"""Analizes a pulse scan There should be at least one channel."""
import sys

import matplotlib.pyplot as plt
import numpy as np
from vdaq_ana import VDaqData
from vdaq_ana import ShowProgress
from vdaq_ana.utils import fit_utils
from vdaq_ana.utils.remove_outliers import remove_outliers

def name2var(name):
    """Convert variable name to variable code."""
    table = {
        "channel": 0,
        "pulse": 1,
        "threshold": 2,
        "trim": 3,
        "bias": 4,
        "xposition": 5,
        "yposition": 6,
        "zposition": 7,
        "angle1": 8,
        "angle2": 9,
        "angle3": 10,
        "user1": 11,
        "user2": 12,
        "user3": 13,
        "user4": 14,
        "thrs_trim": 15,
        "cal_dac": 16,
    }
    return table[name]


def analyze_pulse_scan(fname, options):
    """Main entry."""
    # Open file
    vdaq = VDaqData(fname)
    vdaq.show_info(True)

    scm = vdaq.scan_manager()
    if scm is None:
        print("This is not a scan run.")
        sys.exit()

    print()
    scm.show_info()

    # get the iterator of the Scan data
    prg = ShowProgress(vdaq.nevts)
    keys = list(vdaq.modules.keys())
    if options.mid in keys:
        mid = options.mid
    else:
        mid = keys[0]
    
    md = vdaq.modules[mid]
    point_values = []
    point_mean = []
    point_std = []
    var_id = name2var(options.variable)

    dbg_fig, dbg_ax = plt.subplots(1, 1, tight_layout=True)

    for ipoint, scP in enumerate(vdaq.scan_iter()):
        point_data = []
        for evt in vdaq.create_iterator_at_scanpoint(ipoint, options.mid):
            data = md.process_event(evt)
            if data is not None:
                for C in data:
                    if C.E > options.threshold:
                        point_data.append(C.E)

            prg.increase(show=True)


        if len(point_data) < 100.0:
            continue

        point_values.append(scP.values[var_id])
        point_data = np.array(point_data)
        mean = np.mean(point_data)
        indx = remove_outliers(point_data, 5)
        std = np.std(point_data[indx])

        dbg_ax.clear()
        n, bins, *_ = dbg_ax.hist(point_data, bins=options.nbin, range=(mean-5*std, mean+5*std))
        #result, out, legend = fit_utils.fit_gaussian(n, bins, mean, width=std)
        result, out, legend = fit_utils.fit_two_peaks(n, bins, mean, std/2.0, std)

        point_mean.append(result.best_values["center"])
        point_std.append(result.best_values["sigma"])

        fit_utils.draw_best_fit(dbg_ax, result, bins)
        dbg_ax.legend([legend], loc=1)
        dbg_ax.set_title("Signel Channel signal")
        dbg_ax.set_xlabel("Charge (ADC)")
        plt.draw()
        plt.pause(0.0001)

    fig, ax = plt.subplots(1, 2)
    fig.suptitle("Pulse scan.")
    ax[0].plot(point_values, point_mean, 'o-')
    ax[0].set_title("Signal average")
    ax[0].set_xlabel("Pulse (mV)")

    ax[1].plot(point_values, point_std, 'o-')
    ax[1].set_title("Signal sigma")
    ax[1].set_xlabel("Pulse (mV)")
    
    with open(options.out, "w", encoding="utf-8") as fout:
        fout.write("pulse, ADC\n")
        for V, A in zip(point_values, point_mean):
            fout.write("{}, {}\n".format(V, A))
        
            

    plt.show()


def main():
    """Main entry."""
    from argparse import ArgumentParser

    parser = ArgumentParser()
    parser.add_argument('files', nargs='*', help="Input files")
    parser.add_argument("--mid", dest="mid", default=None, type=int, help="The module ID")
    parser.add_argument("--threshold", default=0.0, type=float, help="Min E to show in histogram")
    parser.add_argument("--nbin", default=50, type=int, help="Number of bins in histogram to fit.")
    parser.add_argument("--out", default="out.csv", help="Output file with gain curve.")
    parser.add_argument("--variable", default="pulse", help="Scan Variable")
    

    opts = parser.parse_args()

    if len(opts.files) == 0:
        print("I need an input file")
        sys.exit()

    analyze_pulse_scan(opts.files[0], opts)
    
if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""Example of data analysis with python readin hdf5 data."""
import os
import sys
import traceback
from argparse import ArgumentParser, Action
import matplotlib.pyplot as plt
import numpy as np
from scipy import stats


try:
    import vdaq_ana

except ImportError:
    import pathlib
    the_folder = pathlib.Path(__file__).parent.parent.parent
    sys.path.append(the_folder.as_posix())

from vdaq_ana.utils import fit_utils
from vdaq_ana import VDaqData, Progress

class CommaSeparatedListAction(Action):
    """Create a list from the comma sepparated numbers at imput."""

    def __call__(self, parser, namespace, values, option_string=None):
        """The actual action."""
        value = {}
        for V in values.split(','):
            try:
                value[int(V)] = 1
            except ValueError:
                if ':' not in V:
                    continue

                items = V.split(':')
                if len(items)==1:
                    continue

                ival = list(map(int, items))
                ival[1] += 1
                for x in range(*ival):
                    value[x] = 1


        setattr(namespace, self.dest, value)


def channel_is_masked(C, chan_list):
    """Checks if the channel is masked"""
    if len(chan_list) == 0:
        return False
    
    for chan in range(C.first, C.last+1):
        if chan in chan_list:
            return True

    return False


def read_data(vdaq, options, evt_E=None, evt_CH=None):
    """Iterates on all the events."""
    if options.nevts is None:
        max_evts = vdaq.nevts
    else:
        max_evts = options.nevts

    prg = Progress.ShowProgress(max_evts)
    prg.start()

    # now we iterate on all the data
    # for that we use vdaq as iterable
    if evt_E is None:
        evt_E = {}

    if evt_CH is None:
        evt_CH = {}

    print("\nIterating over all data in the file")
    ievt = 0
    try:
        for ievt, evt in enumerate(vdaq):
            if ievt == 0:
                continue

            if ievt > max_evts:
                break

            mid = evt.mod_id
            md = vdaq.modules[mid]
            data = md.process_event(evt)
            # get the array to store the event energy
            if data is not None:
                e_list = evt_E.setdefault(mid, [])
                e_ch = evt_CH.setdefault(mid, [])
                for C in data:
                    if channel_is_masked(C, options.mask):
                        continue

                    if C.E > options.threshold:
                        e_list.append(C.E)
                        e_ch.append(C.chan)

            # monitor progress
            prg.increase(show=True)

    except RuntimeError:
        print("\n\nSomething very bad happened...")
        print('-'*60)
        traceback.print_exc(file=sys.stdout)
        print('-'*60)

    except KeyboardInterrupt as exc:
        print("\nArrrrggggg !!!!")
        raise KeyboardInterrupt from exc

    print("\nDone: ", ievt)
    return evt_E, evt_CH


def find_noise_peaks(noise):
    """Find spikes in noise distribution."""
    avg = np.mean(noise)
    sg = np.std(noise)
    peaks = np.where(abs(noise-avg)>3*sg)
    return peaks[0]

def do_analyze_data(file_list, options):
    """Main routine."""

    evt_E = {}
    evt_CH = {}
    for fname in file_list:
        # Open the file
        if not os.path.exists(fname):
            print("Input file", fname, "does not exist")
            continue

        print("\n+++Reading file {}".format(fname))

        # We open here the file
        vdaq = VDaqData(fname)
        vdaq.show_info(True)
        for m in list(vdaq.modules.values()):
            m.print_config()
            m.set_debug(options.debug)
            m.do_cluster = options.cluster
            m.polarity = options.polarity
            
            P = find_noise_peaks(m.noise)
            for P in find_noise_peaks(m.noise):
                options.mask[P] = 1

        # and read the data
        evt_E, evt_CH = read_data(vdaq, options,
                                  evt_E=evt_E,
                                  evt_CH=evt_CH)

    # Create the figure
    n_mod = len(evt_E)
    fig, ax = plt.subplots(nrows=2, ncols=n_mod)
    dax = {}   # Spectrum axes
    daxc = {}  # Hitmap axes
    try:
        for iax, k in enumerate(evt_E.keys()):
            dax[k] = ax[0, iax]
            daxc[k] = ax[1, iax]

    except Exception:
        key = list(evt_E.keys())[0]
        dax[key] = ax[0]
        daxc[key] = ax[1]

    print("")

    # Do the fit
    for mid, E in evt_E.items():
        #  bit of statistics. Find out the optimal binwidth based on average noise
        if options.emin == options.emax:
            vmin = np.min(E)
            vmax = np.max(E)
        else:
            vmin = options.emin
            vmax = options.emax

        noise = stats.mode(vdaq.modules[mid].noise, keepdims=True).mode[0]
        if noise < 5.0:
            noise = 5.0

        if options.nbin < 0:
            rng = vmax - vmin
            nbin = int(rng/(1.0*noise)/5.0 + 0.5) * 5
        else:
            nbin = options.nbin

        # Create and draw the histogram
        ax = dax[mid]
        n, bins, patches = ax.hist(E, nbin, range=(vmin, vmax), facecolor='g', alpha=0.75)
        if options.logy:
            ax.set_yscale("log")

        step = 0.5 * (bins[1] - bins[0])
        X = bins[:-1] + step

        # Do the fits
        if options.fit == "landau":
            mu = np.mean(E)
            sigma = np.std(E)
            result, out, legend = fit_utils.fit_landau_model(mu, sigma, n, bins, options.debug)

        elif options.fit == "gauss":
            result, out, legend = fit_utils.fit_peak_model(n, bins, debug=options.debug)

        if options.fit != "none":
            if options.debug:
                print(result.fit_report())

            fit_utils.draw_best_fit(ax, result, bins)
            ax.legend([legend], loc=1)

        ax.set_title('Module %d' % mid)

        # Hit map
        for mid, CH in evt_CH.items():
            ax = daxc[mid]
            m = vdaq.modules[mid]
            X = np.arange(m.ntot+1)
            n, bins, patches = ax.hist(CH, facecolor='g', bins=m.ntot, alpha=0.75)
            ax.set_xlabel('Channel')
            ax.set_ylabel('Occupancy')

    plt.show()


def analyze_data():
    """Main entry."""
    parser = ArgumentParser()
    parser.add_argument('files', nargs='+', help="Input files")
    parser.add_argument("--fit", choices=["none", "gauss", "landau"],
                        help="It will try a Landau-like fit to the data (default)",
                        default="none"
                        )
    parser.add_argument("--debug", action="store_true", help="Debug", default=False)
    parser.add_argument("--logy", action="store_true", help="Log Scale for spectrum", default=False)
    parser.add_argument("--cluster", action="store_true",
                        help="Will try to find clusters in sparse",
                        default=False)
    parser.add_argument("--nevts", help="Read at most nevts events", default=None, type=int)
    parser.add_argument("--threshold", help="Signal Threshold", default=-9999, type=float)
    parser.add_argument("--polarity", help="Signal polarity", default=1)
    parser.add_argument("--nbin", default=-1, type=int, help="Number of bins in histogram.")
    parser.add_argument("--emin", help="Minimum energy of histogram", default=0, type=float)
    parser.add_argument("--emax", help="maximum energy of histogram", default=0, type=float)
    parser.add_argument("--mask-channels", dest="mask", action=CommaSeparatedListAction, default={},
                        help="Ignore list of channels. The list is made with channel numbers of channel ranges (ch1:ch2 or ch1:ch2:step) ")


    options = parser.parse_args()

    if len(options.files) == 0:
        print("I need an input file")
        sys.exit()

    do_analyze_data(options.files, options)
    sys.exit()


if __name__ == "__main__":
    analyze_data()



#!/usr/bin/env python3
"""Open data file and get spectrum."""
import os
import sys
import runpy
from pathlib import Path
from argparse import Action
from argparse import ArgumentParser

import matplotlib.pyplot as plt
import matplotlib.transforms as transforms
import numpy as np
from scipy.interpolate import CubicSpline
from vdaq_ana.utils.fit_utils import draw_best_fit, fit_gaussian, fit_two_peaks
from vdaq_ana import VDaqData
from vdaq_ana import ShowProgress

from numpy.polynomial.polynomial import polyfit, polyval
from numpy import interp, ndarray, piecewise


def interp1d(x: ndarray, xp, fp):
    """1D piecewise linear interpolation with linear extrapolation."""
    return piecewise(
        x,
        [x < xp[0], (x >= xp[0]) & (x <= xp[-1]), x > xp[-1]],
        [
            lambda xi: polyval(xi, polyfit(xp[:5], fp[:5], 1)),
            lambda xi: interp(xi, xp, fp),
            lambda xi: polyval(xi, polyfit(xp[-5:], fp[-5:], 1)),
        ],
    )


def draw_hist_and_projection(axh, axp, data, title=None, axis_title=None, x_label=None, y_label=None):
    """Plots a bar histogram and the Y projection.

    Args:
        axh: axis for the histogram
        axp: axis for the projection
        data: The monitor data
        title: the histogram title
        axis_title: title for the histogram axis
        x_label: label for the histgram X axis
        y_label: label for the Y axis

    Returns
        mean. std: a tuple with the Y maverage and std

    """
    # Draw the histogram
    y = data
    x = np.linspace(0, len(data)+1, len(data))

    axh.step(x, y)
    if axis_title:
        axh.set_title(axis_title)

    if x_label:
        axh.set_xlabel(x_label)

    if y_label:
        axh.set_ylabel(y_label)

    mean = np.mean(y)
    std = np.std(y)

    # Now the projection
    axp.hist(y, orientation="horizontal")

    return mean, std


def show_pedestals(md):
    """Show module pedestals.

    Args:
        md: Module object.

    """
    # Get the data and plot
    fig = plt.figure(constrained_layout=True)
    gs = fig.add_gridspec(ncols=2, nrows=2, width_ratios=(2, 1))

    # pedestals
    ax = fig.add_subplot(gs[0, 0])
    mped, stped = draw_hist_and_projection(
        ax, fig.add_subplot(gs[0, 1], sharey=ax),
        md.pedestal,
        axis_title="Pedestals",
        x_label="Channel"
    )
    print("Pedestal mean {:.1f} std {:.1f}".format(mped, stped))

    # noise
    ax = fig.add_subplot(gs[1, 0])
    mnoise, stnoise = draw_hist_and_projection(
        ax, fig.add_subplot(gs[1, 1], sharey=ax),
        md.noise,
        axis_title="Noise",
        x_label="Channel"
    )
    print("Noise mean {:.1f} std {:.1f}".format(mnoise, stnoise))



def channel_is_masked(C, chan_list):
    """Checks if the channel is masked"""

    for chan in range(C.first, C.last+1):
        if chan in chan_list:
            return True

    return False

def unity_E(adc, mod=-1):
    """Default calibration function."""
    return adc

def get_calib_func(options):
    """Return calibration function.
        The calibration function accepts the value and an additional parameter
        which will be the module id.
    """
    if not options.calib:
        return unity_E

    fnam = Path(options.calib).expanduser().resolve()
    if not fnam.exists():
        print("Calibration file or function does not exist.")
        return unity_E
    
    try:
        fmod = runpy.run_path(fnam)
        return fmod["get_E"]

    except Exception:
        # We read tehe pairs of adc. E from a file.
        _adc, _E = np.loadtxt(fnam, skiprows=1, delimiter=',', unpack=True)
        cs = CubicSpline(_adc, _E)
        ps = np.polyfit(_adc[:3], _E[:3], 1)
        pf = np.poly1d(ps)

        def get_E(adc, *args):
            """Transform adc into keV"""
            return piecewise(adc,
                            [adc < _adc[0], (adc >= _adc[0]) & (adc < _adc[-1]), adc >= _adc[-1]],
                            [pf, cs, lambda x : np.NaN])

        return get_E

def do_getSpectrum(files, options):
    """Main entry."""

    amplitudes = []
    channels = []

    names = []
    weights = []

    get_E = get_calib_func(options)

    mx_nchan = 0
    for fnam in files:
        # Check that the file exists
        if not os.path.exists(fnam):
            print("Input file", fnam, "does not exist")
            continue

        i1 = fnam.rfind('/')
        i2 = fnam.rfind('_')
        nam = fnam[i1+1:i2]
        names.append(nam)

        # We open here the file with VDaqData
        print("\n\n### Opening {}".format(Path(fnam).name))
        vdaq = VDaqData(fnam)
        vdaq.show_info(True)

        if options.time:
            weights.append(vdaq.get_duration())
        else:
            weights.append(1.0)

        # Find the module
        keys = list(vdaq.modules.keys())
        if options.mid in keys:
            mid = options.mid
        else:
            mid = keys[0]

        print("\n## Looking at data from module {}.".format(mid))
        md = vdaq.modules[mid]
        md.seed_cut = options.seed_cut
        md.neigh_cut = options.neigh_cut
        md.do_cluster = options.cluster
        mx_nchan = max(mx_nchan, md.ntot)

        # Show pedeswtals and noise
        if options.compute_pedestals:
            md.compute_pedestals(options.compute_pedestals)
            
        elif options.load_pedestals:
            md.read_pedestals(options.load_pedestals)
        
        if options.show_ped:
            show_pedestals(md)

        # Create the iterator
        if options.scan_point is not None:
            vdaq_iter = vdaq.create_iterator_at_scanpoint(options.scan_point, [mid])
        elif options.start_time is not None:
            vdaq_iter = vdaq.create_iterator_at_time(options.start_time, [mid])
        elif options.start_evt is not None:
            vdaq_iter = vdaq.create_iterator_at_event(options.start_evt, [mid])
        else:
            vdaq_iter = vdaq.create_module_iterator(md)

        amplitude = []
        chanlist = []
        prg = ShowProgress(vdaq.nevts, width=24)
        prg.start()

        for ievt, evt in enumerate(vdaq_iter):
            data = md.process_event(evt)
            if data is not None:
                for C in data:
                    if channel_is_masked(C, options.mask):
                        continue

                    if C.E > options.threshold and C.E < options.max_E:
                        amplitude.append(C.E)
                        chanlist.append(C.chan)

            prg.increase(show=True)

            if ievt > options.nevt:
                break

        amplitudes.append(get_E(amplitude, mid))
        channels.append(chanlist)
        prg.stop()
        print("")

    # Get the normalization factors
    if len(weights) == 0:
        print("No valid file left for analysis. Quitting")
        sys.exit(-1)

    vmax = np.amax(weights)
    weights = [vmax/W for W in weights]

    # Draw the signal
    if options.calib and options.show_calib:
        fig, ax = plt.subplots(1, 1)
        xxx = np.linspace(0, 1750, int(1750/2.5))
        yyy = get_E(xxx, mid)
        ax.plot(xxx, yyy)
        ax.grid()
        ax.set_xlabel("ADC counts")
        ax.set_ylabel("Energy (keV)")

    fig, ax = plt.subplots(1, 1)
    i = 0
    ymax = -1e100
    for A, nam in zip(amplitudes, names):
        n, bins, P = ax.hist(A, bins=options.nbin,
                             histtype='stepfilled', alpha=0.25, label=nam)

        if options.time:
            W = weights[i]
            for p in P:
                points = p.get_xy()
                new_points = np.zeros(points.shape)
                ipoint = 0
                for x in points:
                    y = W*x[1]
                    ymax = max(y, ymax)
                    new_points[ipoint, :] = [x[0], y]
                    ipoint += 1

                p.set_xy(new_points)

            i += 1

    if options.time:
        y0, y1 = ax.get_ylim()
        ax.set_ylim(y0, 1.1*ymax)

    # n, bins, *_ = ax.hist(amplitudes, bins=options.nbin, label=names)
    legends = []
    if options.fit:
        for amplitude in amplitudes:
            mean = np.mean(amplitude)
            std = np.std(amplitude)
            if options.two_peaks:
                result, out, legend = fit_two_peaks(mean, std, std, n, bins)

            else:
                result, out, legend = fit_gaussian(n, bins, mean, width=std)

            draw_best_fit(ax, result, bins)
            legends.append(legend)

        ax.legend(legends, loc=1)
    else:
        ax.legend(names, loc=1)

    ax.grid()
    ax.set_title("Signal")
    if options.calib:
        ax.set_xlabel("Energy (keV)")
    else:
        ax.set_xlabel("Charge (ADC)")

    if options.logY:
        ax.set_yscale("log")

    if options.hitmap:
        fig, ax = plt.subplots(1, 1)
        i = 0
        ymax = -1e100
        for A, nam in zip(channels, names):
            n, bins, P = ax.hist(A, bins=int(mx_nchan),
                                 histtype='stepfilled', alpha=0.25, label=nam)
        ax.set_xlabel("channel number")
        ax.set_ylabel("Number  of hits")
        ax.legend()

    plt.show()


class CommaSeparatedListAction(Action):
    """Create a list from the comma sepparated numbers at imput."""

    def __call__(self, parser, namespace, values, option_string=None):
        """The actual action."""
        setattr(namespace, self.dest, list(map(int, values.split(','))))


class ChannelMaskAction(Action):
    """Create a list from the comma sepparated numbers at imput."""

    def __call__(self, parser, namespace, values, option_string=None):
        """The actual action."""
        value = {}
        for V in values.split(','):
            try:
                value[int(V)] = 1
            except ValueError:
                if ':' not in V:
                    continue

                items = V.split(':')
                if len(items)==1:
                    continue

                ival = list(map(int, items))
                ival[1] += 1
                for x in range(*ival):
                    value[x] = 1


        setattr(namespace, self.dest, value)


def getSpectrum():
    """Main entry."""

    parser = ArgumentParser()
    parser.add_argument('files', nargs='+', help="Input files")
    parser.add_argument("--nevt", default=sys.maxsize, type=int, help="max number of events.")
    parser.add_argument("--start-time", dest="start_time", default=None, type=float,
                        help="Event time to start")
    parser.add_argument("--start-evt", dest="start_evt", default=None,
                        action=CommaSeparatedListAction,
                        help="Event number to start showing")
    parser.add_argument("--scan-point", dest="scan_point", type=int, default=None,
                        help="Scan point number to visit")
    parser.add_argument("--compute_pedestals", default=None, type=str, help="File name to store computed pedestals.")
    parser.add_argument("--load_pedestals", default=None, type=str, help="File name to read pedestal values.")
    
    parser.add_argument("--nbin", default=50, type=int, help="Number of bins in histogram.")
    parser.add_argument("--logY", default=False, action="store_true", help="Log axis")
    parser.add_argument("--threshold", default=0.0, type=float, help="Min E to show in histogram")
    parser.add_argument("--max_E", default=sys.float_info.max, type=float, help="Max E to show in histogram")
    parser.add_argument("--two_peaks", default=False, action="store_true", help="Do a 2-peak fit")
    parser.add_argument("--no-fit", dest="fit", default=True, action="store_false")
    parser.add_argument("--show-ped", action="store_true", default=False, dest="show_ped")
    parser.add_argument("--hitmap", action="store_true", default=False, help="Show hitmap")
    parser.add_argument("--mid", dest="mid", default=-1, type=int, help="The module ID")
    parser.add_argument("--calib", default=None, help="The energy calibration file. X:adc, Y:energy. CAn also be a python file implementing the get_E function.")
    parser.add_argument("--show_calib", default=False, action="store_true", help="Show the calibration curve")
    
    parser.add_argument("--time", default=False, action="store_true",
                        help="Normalize to duration of run when reading various files")
    parser.add_argument("--seed_cut", dest="seed_cut", default=3.5, type=float,
                        help="Threshold of charge for a cluster seed")
    parser.add_argument("--neigh_cut", dest="neigh_cut", default=1.5, type=float,
                        help="Threshold of charge to add a neighbour to a cluster")
    parser.add_argument("--cluster", action="store_true",
                        help="Will try to find clusters in sparse",
                        default=False)

    parser.add_argument("--mask-channels", dest="mask", action=ChannelMaskAction, default={},
                        help="Ignore list of channels. The list is made with channel numbers of channel ranges (ch1:ch2 or ch1:ch2:step) ")

    options = parser.parse_args()

    if len(options.files) == 0:
        print("I need an input file")
        sys.exit()

    do_getSpectrum(options.files, options)


if __name__ == "__main__":
    getSpectrum()



#!/usr/bin/env python3
""" Read data and cluster hits in time.
"""
import sys
from pathlib import Path
from vdaq_ana import VDaqData, Progress
from vdaq_ana.utils import fit_utils

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors

def main(fname, options):
    """ Main entry.

    Args:
        fname: the input filename
        options: the program options

    """
    inp_file_path = Path(fname)
    if not inp_file_path.exists():
        print("Input file ", inp_file_path, " does not exist")
        return

    # open the vdaq file
    vdaq = VDaqData(fname)
    vdaq.show_info()

    # setup the progress monitor
    if options.nevts is None:
        max_evts = vdaq.nevts
    else:
        max_evts = options.nevts

    prg = Progress.ShowProgress(max_evts)

    # This is to stor e theTDC of previous event
    last_tdc = {}
    for m in list(vdaq.modules.values()):
        last_tdc[m.id] = 0.0

    # Get events sorted by TDC
    vdaq.iter_key = 'tdc'

    deltaT = []
    all_events_E = []
    all_events_C = []
    all_events_dt = []
    isolated_events_E = []
    isolated_events_C = []
    prg.start()
    for ievt, evt in enumerate(vdaq):
        # monitor progress
        prg.increase(show=True)

        if ievt > max_evts:
            break

        mid = evt.mod_id
        ltdc = last_tdc.get(mid, -1)
        last_tdc[mid] = evt.tdc

        # time distnce with previous event
        dt = evt.tdc - ltdc

        # get the module
        module = vdaq.modules[mid]

        if dt > 1e5:
            continue

        if ievt < 3:
            continue

        deltaT.append(dt)

        # Analysis
        data = module.process_event(evt)
        if data is not None:
            for channel, E in data:
                if E > options.threshold:
                    all_events_E.append(E)
                    all_events_dt.append(dt)
                    all_events_C.append(channel)

        if dt > options.dtmin:
            for channel, E in data:
                if E > options.threshold:
                    isolated_events_E.append(E)
                    isolated_events_C.append(channel)


    prg.stop()
    print("")

    fig = plt.figure(figsize=[5, 8],tight_layout=True)
    fig.subplots_adjust(left=0.05, right=0.95)

    ax = fig.add_subplot(3,1,1)
    ax.set_title("Event time separation")
    ax.set_yscale("log")
    ax.set_xscale("log")

    ax.hist(deltaT, bins=10**np.arange(0,4,0.05))
    ax.set_xlabel("time sep (ns)")

    ax = fig.add_subplot(3,2,3)
    fig.suptitle(inp_file_path.name)
    n, bins, patches = ax.hist(all_events_E, bins=150)
    ax.set_title("All events")
    ax.set_xlabel("Energy (ADC)")
    if options.fit:
        imax = np.argmax(n)
        mu = bins[imax]
        sigma = np.std(all_events_E)
        print("µ {:.1f} - RMS {:.1f}".format(mu, sigma))
        result, out, legend = fit_utils.fit_landau_model(mu, sigma, n, bins)
        fit_utils.draw_best_fit(ax, result, bins)
        ax.legend([legend], loc=1)


    ax = fig.add_subplot(3,2,4)
    n, bins, patches = ax.hist(isolated_events_E, bins=150)
    ax.set_title("Isolated events\n dt>{}".format(options.dtmin))
    ax.set_xlabel("Energy (ADC)")

    ax = fig.add_subplot(3, 1, 3)
    h = ax.hist2d(all_events_E, all_events_dt,
                  bins=[50, 10**np.arange(0,4,0.05)],
                  cmap=plt.cm.jet,
                  norm=matplotlib.colors.LogNorm())
    fig.colorbar(h[3], ax=ax)
    ax.set_yscale("log")
    ax.set_xlabel("Energy (ADC)")
    ax.set_ylabel("Time sep (ns)")

    plt.show()

if __name__ == "__main__":
    from argparse import ArgumentParser
    parser = ArgumentParser()
    parser.add_argument('files', nargs='*', help="Input files")
    parser.add_argument("--nevts", help="Read at most nevts events", default=None)
    parser.add_argument("--fit", action="store_true",
                        help="It will try a Landau-like fit to the data (default)",
                        default=False
                        )
    parser.add_argument("--threshold", help="Signal Threshold", default=-9999, type=float)
    parser.add_argument("--dtmin", help="Min time distance", default=1e5, type=float)

    opt = parser.parse_args()
    if len(opt.files) == 0:
        print(sys.argv[0])
        print("I need an input file")
        sys.exit()

    main(opt.files[0], opt)
